docker run --rm -v`pwd`:/local -v`realpath install-python-manylinux2014-x64`:/casadi_python:ro -v`realpath install-octave-manylinux2014-x64/casadi`:/casadi_octave:ro --env PYTHONPATH=/casadi_python --env OCTAVE_PATH=/casadi_octave --env LD_LIBRARY_PATH=/casadi_python/casadi -u $(id -u):$(id -g)  ghcr.io/casadi/ci-doc:latest /bin/bash -c "cd misc && python add_identifiers.py && cd .. && cd docs/users_guide && rm -rf build && rm -f snippets/* && make users_guide.pdf && cd ../../test && make user_guide_snippets_py user_guide_snippets_oct && cd .. && rm -rf docs/users_guide/build && cd docs/users_guide && make users_guide.pdf bare_toc bare_body && cd ../.. && cd docs/api && make full doxexport"
cd web && zip -rq ../website_assets.zip *
