LIBRARY libcasadi.dll
EXPORTS
    casadi_c_activate
    casadi_c_checkout
    casadi_c_checkout_id
    casadi_c_clear
    casadi_c_decref
    casadi_c_decref_id
    casadi_c_default_in
    casadi_c_default_in_id
    casadi_c_eval
    casadi_c_eval_id
    casadi_c_id
    casadi_c_incref
    casadi_c_incref_id
    casadi_c_int_width
    casadi_c_name
    casadi_c_name_id
    casadi_c_name_in
    casadi_c_name_in_id
    casadi_c_name_out
    casadi_c_name_out_id
    casadi_c_n_in
    casadi_c_n_in_id
    casadi_c_n_loaded
    casadi_c_n_out
    casadi_c_n_out_id
    casadi_c_pop
    casadi_c_push_file
    casadi_c_real_width
    casadi_c_release
    casadi_c_release_id
    casadi_c_sparsity_in
    casadi_c_sparsity_in_id
    casadi_c_sparsity_out
    casadi_c_sparsity_out_id
    casadi_c_work
    casadi_c_work_id
